#!/usr/bin/env python3

import argparse
import dbus

NAME = 'org.gnome.Mutter.DebugControl'
INTERFACE = 'org.gnome.Mutter.DebugControl'
OBJECT_PATH = '/org/gnome/Mutter/DebugControl'

PROPS_IFACE = 'org.freedesktop.DBus.Properties'

def bool_to_string(value):
    if value:
        return "true"
    else:
        return "false"

def string_to_bool(value):
    if value == "true":
        return True
    if value == "false":
        return False
    raise BaseException(f"bad boolean value: {value}")

def value_to_string(value):
    if isinstance(value, dbus.Boolean):
        return bool_to_string(value)
    return f"{value}"

def string_to_value(current, value):
    if isinstance(current, dbus.Boolean):
        return dbus.Boolean(string_to_bool(value), variant_level=1)
    if isinstance(current, dbus.UInt32):
        return dbus.UInt32(int(value), variant_level=1)
    return value

def get_debug_control():
    bus = dbus.SessionBus()
    try:
        debug_control = bus.get_object(NAME, OBJECT_PATH)
    except dbus.exceptions.DBusException:
        print("The DebugControl service is not available.")
        print("You may have to enable the `debug-control` flag in looking glass.")
        exit(-1)
    return debug_control

def status():
    debug_control = get_debug_control()
    props = debug_control.GetAll(INTERFACE, dbus_interface=PROPS_IFACE)
    for prop, value in props.items():
        value = value_to_string (value)
        print(f"{prop}: {value}")

def enable(prop):
    debug_control = get_debug_control()
    debug_control.Set(INTERFACE, prop, dbus.Boolean(True, variant_level=1),
                      dbus_interface=PROPS_IFACE)

def disable(prop):
    debug_control = get_debug_control()
    debug_control.Set(INTERFACE, prop, dbus.Boolean(False, variant_level=1),
                      dbus_interface=PROPS_IFACE)

def toggle(prop):
    debug_control = get_debug_control()

    value = debug_control.Get(INTERFACE, prop, dbus_interface=PROPS_IFACE)
    debug_control.Set(INTERFACE, prop, dbus.Boolean(not value, variant_level=1),
                      dbus_interface=PROPS_IFACE)

def set_value(kv):
    debug_control = get_debug_control()
    [prop, value] = kv

    current = debug_control.Get(INTERFACE, prop, dbus_interface=PROPS_IFACE)
    value = string_to_value (current, value)

    debug_control.Set(INTERFACE, prop, value, dbus_interface=PROPS_IFACE)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Get and set debug state')

    parser.add_argument('--status', action='store_true')
    parser.add_argument('--enable', metavar='PROPERTY', type=str, nargs='?')
    parser.add_argument('--disable', metavar='PROPERTY', type=str, nargs='?')
    parser.add_argument('--toggle', metavar='PROPERTY', type=str, nargs='?')
    parser.add_argument('--set', metavar=('PROPERTY', 'VALUE'), type=str, nargs=2)

    args = parser.parse_args()
    if args.status:
        status()
    elif args.enable:
        enable(args.enable)
    elif args.disable:
        disable(args.disable)
    elif args.toggle:
        toggle(args.toggle)
    elif args.set:
        set_value(args.set)
    else:
        parser.print_usage()



#!/usr/bin/env python3

import argparse
import enum
import subprocess
import sys

from gi.repository import GLib, Gio


NAME = 'org.gnome.Mutter.DisplayConfig'
INTERFACE = 'org.gnome.Mutter.DisplayConfig'
OBJECT_PATH = '/org/gnome/Mutter/DisplayConfig'

TRANSFORM_STRINGS = {
    0: 'normal',
    1: '90',
    2: '180',
    3: '270',
    4: 'flipped',
    5: 'flipped-90',
    6: 'flipped-180',
    7: 'flipped-270',
}

COLOR_MODES = {
    0: 'default',
    1: 'BT.2100',
}


class Source(enum.Enum):
    DBUS = 1
    COMMAND_LINE = 2
    FILE = 3


class MonitorConfig:
    CONFIG_VARIANT_TYPE = GLib.VariantType.new(
        '(ua((ssss)a(siiddada{sv})a{sv})a(iiduba(ssss)a{sv})a{sv})')

    def get_current_state(self) -> GLib.Variant:
        raise NotImplementedError()

    def parse_data(self):
        """TODO: add data parser so that can be used for reconfiguring"""

    def print_data(self, *, level, is_last, lines, data):
        if is_last:
            link = '└'
        else:
            link = '├'
        padding = ' '

        if level >= 0:
            indent = level
            buffer = f'{link:{padding}>{indent * 4}}──{data}'
            buffer = list(buffer)
            for line in lines:
                if line == level:
                    continue
                index = line * 4
                if line > 0:
                    index -= 1
                buffer[index] = '│'
            buffer = ''.join(buffer)
        else:
            buffer = data

        print(buffer)

        if is_last and level in lines:
            lines.remove(level)
        elif not is_last and level not in lines:
            lines.append(level)

    def maybe_describe(self, property, value):
        if property == 'color-modes':
            return [COLOR_MODES.get(color_mode)
                    for color_mode in value]
        else:
            return value

    def print_properties(self, *, level, lines, properties):
        property_list = list(properties)

        self.print_data(level=level, is_last=True, lines=lines,
                        data=f'Properties: ({len(property_list)})')
        for property in property_list:
            is_last = property == property_list[-1]
            property_value = self.maybe_describe(property, properties[property])
            self.print_data(level=level + 1, is_last=is_last, lines=lines,
                            data=f'{property} ⇒ {property_value}')

    def print_current_state(self, short):
        variant = self.get_current_state()

        print('Serial: {}'.format(variant[0]))
        print()
        print('Monitors:')
        monitors = variant[1]
        lines = []
        for monitor in monitors:
            is_last = monitor == monitors[-1]
            spec = monitor[0]
            modes = monitor[1]
            properties = monitor[2]
            self.print_data(level=0, is_last=is_last, lines=lines,
                            data='Monitor {}'.format(spec[0]))
            self.print_data(level=1, is_last=False, lines=lines,
                            data=f'EDID: vendor: {spec[1]}, product: {spec[2]}, serial: {spec[3]}')

            mode_count = len(modes)
            if short:
                modes = [mode for mode in modes if len(mode[6]) > 0]
                self.print_data(level=1, is_last=False, lines=lines,
                                data=f'Modes ({len(modes)}, {mode_count - len(modes)} omitted)')
            else:
                self.print_data(level=1, is_last=False, lines=lines,
                                data=f'Modes ({len(modes)})')

            for mode in modes:
                is_last = mode == modes[-1]
                self.print_data(level=2, is_last=is_last, lines=lines,
                                data=f'{mode[0]}')
                self.print_data(level=3, is_last=False, lines=lines,
                                data=f'Dimension: {mode[1]}x{mode[2]}')
                self.print_data(level=3, is_last=False, lines=lines,
                                data=f'Refresh rate: {mode[3]:.3f}')
                self.print_data(level=3, is_last=False, lines=lines,
                                data=f'Preferred scale: {mode[4]}')
                self.print_data(level=3, is_last=False, lines=lines,
                                data=f'Supported scales: {mode[5]}')

                mode_properties = mode[6]
                self.print_properties(level=3, lines=lines,
                                      properties=mode_properties)

            self.print_properties(level=1, lines=lines, properties=properties)

        print()
        print('Logical monitors:')
        logical_monitors = variant[2]
        index = 1
        for logical_monitor in logical_monitors:
            is_last = logical_monitor == logical_monitors[-1]
            properties = logical_monitor[2]
            self.print_data(level=0, is_last=is_last, lines=lines,
                            data=f'Logical monitor #{index}')
            self.print_data(level=1, is_last=False, lines=lines,
                            data=f'Position: ({logical_monitor[0]}, {logical_monitor[1]})')
            self.print_data(level=1, is_last=False, lines=lines,
                            data=f'Scale: {logical_monitor[2]}')
            self.print_data(level=1, is_last=False, lines=lines,
                            data=f'Transform: {TRANSFORM_STRINGS.get(logical_monitor[3])}')
            self.print_data(level=1, is_last=False, lines=lines,
                            data=f'Primary: {logical_monitor[4]}')
            monitors = logical_monitor[5]
            self.print_data(level=1, is_last=False, lines=lines,
                            data=f'Monitors: ({len(monitors)})')
            for monitor in monitors:
                is_last = monitor == monitors[-1]
                self.print_data(level=2, is_last=is_last, lines=lines,
                                data=f'{monitor[0]} ({monitor[1]}, {monitor[2]}, {monitor[3]})')

            properties = logical_monitor[6]
            self.print_properties(level=1, lines=lines, properties=properties)

            index += 1

        properties = variant[3]
        print()
        self.print_properties(level=-1, lines=lines, properties=properties)


class MonitorConfigDBus(MonitorConfig):
    def __init__(self):
        self._proxy = Gio.DBusProxy.new_for_bus_sync(
            bus_type=Gio.BusType.SESSION,
            flags=Gio.DBusProxyFlags.NONE,
            info=None,
            name=NAME,
            object_path=OBJECT_PATH,
            interface_name=INTERFACE,
            cancellable=None,
        )

    def get_current_state(self) -> GLib.Variant:
        variant = self._proxy.call_sync(
            method_name='GetCurrentState',
            parameters=None,
            flags=Gio.DBusCallFlags.NO_AUTO_START,
            timeout_msec=-1,
            cancellable=None
        )
        assert variant.get_type().equal(self.CONFIG_VARIANT_TYPE)
        return variant


class MonitorConfigCommandLine(MonitorConfig):
    def get_current_state(self) -> GLib.Variant:
        command = ('gdbus call -e '
                   f'-d {NAME} '
                   f'-o {OBJECT_PATH} '
                   f'-m {INTERFACE}.GetCurrentState')

        result = subprocess.run(command, shell=True,
                                check=True, capture_output=True, text=True)
        return GLib.variant_parse(self.CONFIG_VARIANT_TYPE, result.stdout)


class MonitorConfigFile(MonitorConfig):
    def __init__(self, file_path):
        if file_path == '-':
            self._data = sys.stdin.read()
        else:
            with open(file_path) as file:
                self._data = file.read()

    def get_current_state(self) -> GLib.Variant:
        return GLib.variant_parse(self.CONFIG_VARIANT_TYPE, self._data)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Get display state')
    parser.add_argument('--file', metavar='FILE', type=str, nargs='?',
                        help='Read the output from gdbus call instead of calling D-Bus')
    parser.add_argument('--gdbus', action='store_true')
    parser.add_argument('--short', action='store_true')

    args = parser.parse_args()

    if args.file and args.gdbus:
        raise argparse.ArgumentTypeError('Incompatible arguments')

    if args.file:
        monitor_config = MonitorConfigFile(args.file)
    elif args.gdbus:
        monitor_config = MonitorConfigCommandLine()
    else:
        monitor_config = MonitorConfigDBus()

    monitor_config.print_current_state(short=args.short)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of mutter.
#
# Copyright (C) 2024 Alberto Ruiz <aruiz@gnome.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Alberto Ruiz <aruiz@gnome.org>

# Python port of <https://gitlab.gnome.org/GNOME/gtk/-/blob/main/gdk/gen-keyname-table.pl>

from dataclasses import dataclass
import sys
import re
import time


@dataclass
class Key:
    name: str
    keyval: int
    offset: int = 0


if __name__ == "__main__":
    if len(sys.argv) != 2:
        sys.exit(f"Usage: {sys.argv[0]} keynames.txt > keyname-table.h")

    keys = []
    with open(sys.argv[1], "r", encoding="utf8") as keynames:
        for line in keynames.readlines():
            if line.startswith("!"):
                continue

            match = re.match(r"^\s*(0x[0-9a-f]+)\s+([\w_]*\S)\s+(1)?\s*$", line)
            if not match:
                sys.exit(f"Failed to parse line: {line}")

            groups = match.groups()
            keys.append(Key(groups[1], int(groups[0])))

    formatted_time = time.strftime("%a %b %d %H:%M:%S %Y", time.gmtime())
    print(f"""/* clutter-keyname-table.h: Generated by gen-keyname-table.py from keynames.txt
 *
 *  Date: {formatted_time}
 *
 * Do not edit.
 */
static const char keynames[] =""")

    offset = 0
    for key in keys:
        if offset != 0:
            print("")

        print(f'  "{key.name}\\0"', end="")

        key.offset = offset
        offset += len(key.name) + 1

    print(";")

    print("""
typedef struct {
    unsigned int keyval;
    unsigned int offset;
} clutter_key;

static const clutter_key clutter_keys_by_keyval[] = {""")

    for i, key in enumerate(keys):
        if i != 0:
            print(",")

        print("".join(f"  {{ {key.keyval}, {key.offset} }}"), end="")

    print("\n};\n")



'''colord proxy mock template
'''

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = 'Jonas Ådahl'
__copyright__ = '(c) 2021 Red Hat Inc.'

import dbus
import os
import pwd
from dbusmock import MOCK_IFACE, mockobject


BUS_PREFIX = 'org.freedesktop.ColorManager'
PATH_PREFIX = '/org/freedesktop/ColorManager'

BUS_NAME = BUS_PREFIX
MAIN_OBJ = PATH_PREFIX
MAIN_IFACE = BUS_NAME
DEVICE_IFACE = BUS_PREFIX + '.Device'
PROFILE_IFACE = BUS_PREFIX + '.Profile'
SYSTEM_BUS = True


def load(mock, parameters=None):
    mock.devices = {}
    mock.profiles = {}

def escape_unit_name(name):
    for s in ['.', '-', '\'', ' ']:
        name = name.replace(s, '_')
    return name

def get_username(uid):
    return pwd.getpwuid(uid).pw_name

def device_id_from_path(mock, path):
    for device_id in mock.devices:
        device_path = mock.devices[device_id]
        if device_path == path:
            return device_id
    return None

def profile_id_from_path(mock, path):
  for profile_id in mock.profiles:
     profile_path = mock.profiles[profile_id]
     if profile_path == path:
         return profile_id
  return None


class ColordAlreadyExistsException(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.ColorManager.AlreadyExists'

    def __init__(self, message=''):
        super().__init__(message)

class ColordNotFoundException(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.ColorManager.NotFound'


@dbus.service.method(MAIN_IFACE, in_signature='ssa{sv}', out_signature='o')
def CreateDevice(self, device_id, scope, props):
    if device_id in self.devices:
        raise ColordAlreadyExistsException(
            f"device id '{device_id}' already exists")

    uid = os.getuid()
    username = get_username(uid)
    device_path = PATH_PREFIX + '/devices/' + \
        escape_unit_name(device_id) + \
        '_' + escape_unit_name(username) + '_' + str(uid)
    self.devices[device_id] = device_path
    self.AddObject(device_path,
                   DEVICE_IFACE,
                   {
                     'DeviceId': device_id,
                     'Profiles': dbus.types.Array(signature='o'),
                     'Enabled': True,
                     'ProfilingInhibitors': dbus.types.Array(signature='s'),
                   },
                   [])
    self.EmitSignal(MAIN_IFACE, 'DeviceAdded', 'o', [device_path])
    return device_path

@dbus.service.method(MAIN_IFACE, in_signature='o')
def DeleteDevice(self, device_path):
    self.RemoveObject(device_path)
    device_id = device_id_from_path(self, device_path)
    del self.devices[device_id]
    self.EmitSignal(MAIN_IFACE, 'DeviceRemoved', 'o', [device_path])


@dbus.service.method(MAIN_IFACE, in_signature='s', out_signature='o')
def FindDeviceById(self, device_id):
    try:
        return self.devices[device_id]
    except KeyError:
        raise dbus.exceptions.DBusException(
            f"Device {device_id} not found",
            name="org.freedesktop.ColorManager.NotFound"
        )


@dbus.service.method(MAIN_IFACE, in_signature='ssha{sv}', out_signature='o')
def CreateProfileWithFd(self, profile_id, scope, handle, props):
    uid = os.getuid()
    username = get_username(uid)
    profile_path = PATH_PREFIX + '/profiles/' + \
        escape_unit_name(profile_id) + \
        '_' + escape_unit_name(username) + '_' + str(uid)

    if profile_id in self.profiles:
        raise ColordAlreadyExistsException(
            f"profile id '{profile_id}' already exists")

    self.profiles[profile_id] = profile_path
    self.AddObject(profile_path,
                   PROFILE_IFACE,
                   {
                     'ProfileId': profile_id,
                     'Enabled': True,
                     'Filename': props['Filename'],
                   },
                   [])
    self.EmitSignal(MAIN_IFACE, 'ProfileAdded', 'o', [profile_path])
    return profile_path

@dbus.service.method(MAIN_IFACE, in_signature='o')
def DeleteProfile(self, profile_path):
    self.RemoveObject(profile_path)
    profile_id = profile_id_from_path(self, profile_path)
    del self.profiles[profile_id]
    self.EmitSignal(MAIN_IFACE, 'ProfileRemoved', 'o', [profile_path])


@dbus.service.method(MAIN_IFACE, in_signature='s', out_signature='o')
def FindProfileById(self, profile_id):
    if profile_id in self.devices:
        return self.devices[profile_id]
    else:
        raise ColordNotFoundException()


@dbus.service.method(MOCK_IFACE)
def Reset(self):
    for device_path in self.devices.values():
        self.RemoveObject(device_path)
    self.devices = {}
    for profile_path in self.profiles.values():
        self.RemoveObject(profile_path)
    self.profiles = {}

@dbus.service.method(MOCK_IFACE, in_signature='ss')
def AddSystemProfile(self, profile_id, file_path):
    profile_path = PATH_PREFIX + '/profiles/' + \
        escape_unit_name(profile_id)
    self.profiles[profile_id] = profile_path
    self.AddObject(profile_path,
                   PROFILE_IFACE,
                   {
                     'ProfileId': profile_id,
                     'Filename': file_path,
                     'Enabled': True,
                   },
                   [])
    self.EmitSignal(MAIN_IFACE, 'ProfileAdded', 'o', [profile_path])

@dbus.service.method(MOCK_IFACE, in_signature='sas')
def SetDeviceProfiles(self, device_id, profile_ids):
    device_path = self.devices[device_id]
    device = mockobject.objects[device_path]
    profile_paths = [
        dbus.types.ObjectPath(self.profiles[profile_id])
        for profile_id in profile_ids
    ]
    device.UpdateProperties(DEVICE_IFACE, {'Profiles': dbus.types.Array(profile_paths)})
    device.EmitSignal(DEVICE_IFACE, 'Changed', '', [])



'''gsd-color proxy mock template
'''

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = 'Jonas Ådahl'
__copyright__ = '(c) 2021 Red Hat Inc.'

import dbus
import os
from dbusmock import MOCK_IFACE


BUS_NAME = 'org.gnome.SettingsDaemon.Color'
MAIN_OBJ = '/org/gnome/SettingsDaemon/Color'
MAIN_IFACE = BUS_NAME
SYSTEM_BUS = False


def load(mock, parameters=None):
    mock.night_light_active = False
    mock.temperature = 6500

    mock.AddProperty(MAIN_IFACE, 'NightLightActive', dbus.Boolean())
    mock.AddProperty(MAIN_IFACE, 'Temperature', dbus.UInt32())
    mock.Set(MAIN_IFACE, 'NightLightActive', mock.night_light_active)
    mock.Set(MAIN_IFACE, 'Temperature', dbus.UInt32(mock.temperature))


@dbus.service.method(MOCK_IFACE, in_signature='b')
def SetNightLightActive(self, active):
    self.UpdateProperties(MAIN_IFACE, {'NightLightActive': active})

@dbus.service.method(MOCK_IFACE, in_signature='u')
def SetTemperature(self, temperature):
    self.UpdateProperties(MAIN_IFACE, {'Temperature': temperature})
